@@ -16,17 +16,18 @@
 
 package com.android.apksigner;
 
+import static com.android.apksig.Constants.LIBRARY_PAGE_ALIGNMENT_BYTES;
+
 import com.android.apksig.ApkSigner;
 import com.android.apksig.ApkVerifier;
 import com.android.apksig.SigningCertificateLineage;
 import com.android.apksig.SigningCertificateLineage.SignerCapabilities;
 import com.android.apksig.apk.ApkFormatException;
 import com.android.apksig.apk.MinSdkVersionException;
+import com.android.apksig.internal.apk.v3.V3SchemeConstants;
 import com.android.apksig.util.DataSource;
 import com.android.apksig.util.DataSources;
 
-import org.conscrypt.OpenSSLProvider;
-
 import java.io.BufferedReader;
 import java.io.File;
 import java.io.IOException;
@@ -50,6 +51,7 @@ import java.security.interfaces.ECKey;
 import java.security.interfaces.RSAKey;
 import java.util.ArrayList;
 import java.util.Arrays;
+import java.util.Base64;
 import java.util.List;
 
 /**
@@ -64,6 +66,8 @@ public class ApkSignerTool {
     private static final String HELP_PAGE_VERIFY = "help_verify.txt";
     private static final String HELP_PAGE_ROTATE = "help_rotate.txt";
     private static final String HELP_PAGE_LINEAGE = "help_lineage.txt";
+    private static final String BEGIN_CERTIFICATE = "-----BEGIN CERTIFICATE-----";
+    private static final String END_CERTIFICATE = "-----END CERTIFICATE-----";
 
     private static MessageDigest sha256 = null;
     private static MessageDigest sha1 = null;
@@ -80,7 +84,9 @@ public class ApkSignerTool {
             return;
         }
 
+        // BEGIN-AOSP
         addProviders();
+        // END-AOSP
 
         String cmd = params[0];
         try {
@@ -113,18 +119,20 @@ public class ApkSignerTool {
         }
     }
 
+    // BEGIN-AOSP
     /**
      * Adds additional security providers to add support for signature algorithms not covered by
      * the default providers.
      */
     private static void addProviders() {
         try {
-            Security.addProvider(new OpenSSLProvider());
+            Security.addProvider(new org.conscrypt.OpenSSLProvider());
         } catch (UnsatisfiedLinkError e) {
             // This is expected if the library path does not include the native conscrypt library;
             // the default providers support all but PSS algorithms.
         }
     }
+    // END-AOSP
 
     private static void sign(String[] params) throws Exception {
         if (params.length == 0) {
@@ -140,11 +148,17 @@ public class ApkSignerTool {
         boolean v3SigningEnabled = true;
         boolean v4SigningEnabled = true;
         boolean forceSourceStampOverwrite = false;
+        boolean sourceStampTimestampEnabled = true;
+        boolean alignFileSize = false;
         boolean verityEnabled = false;
         boolean debuggableApkPermitted = true;
+        boolean alignmentPreserved = false;
+        int libPageAlignment = LIBRARY_PAGE_ALIGNMENT_BYTES;
         int minSdkVersion = 1;
         boolean minSdkVersionSpecified = false;
         int maxSdkVersion = Integer.MAX_VALUE;
+        int rotationMinSdkVersion = V3SchemeConstants.DEFAULT_ROTATION_MIN_SDK_VERSION;
+        boolean rotationTargetsDevRelease = false;
         List<SignerParams> signers = new ArrayList<>(1);
         SignerParams signerParams = new SignerParams();
         SigningCertificateLineage lineage = null;
@@ -157,6 +171,8 @@ public class ApkSignerTool {
         String optionOriginalForm = null;
         boolean v4SigningFlagFound = false;
         boolean sourceStampFlagFound = false;
+        boolean deterministicDsaSigning = false;
+        boolean otherSignersSignaturesPreserved = false;
         while ((optionName = optionsParser.nextOption()) != null) {
             optionOriginalForm = optionsParser.getOptionOriginalForm();
             if (("help".equals(optionName)) || ("h".equals(optionName))) {
@@ -171,7 +187,13 @@ public class ApkSignerTool {
                 minSdkVersionSpecified = true;
             } else if ("max-sdk-version".equals(optionName)) {
                 maxSdkVersion = optionsParser.getRequiredIntValue("Maximum API Level");
-            } else if ("v1-signing-enabled".equals(optionName)) {
+            } else if ("rotation-min-sdk-version".equals(optionName)) {
+                rotationMinSdkVersion = optionsParser.getRequiredIntValue(
+                        "Minimum API Level for Rotation");
+            } else if ("rotation-targets-dev-release".equals(optionName)) {
+                rotationTargetsDevRelease = optionsParser.getOptionalBooleanValue(true);
+            }
+            else if ("v1-signing-enabled".equals(optionName)) {
                 v1SigningEnabled = optionsParser.getOptionalBooleanValue(true);
             } else if ("v2-signing-enabled".equals(optionName)) {
                 v2SigningEnabled = optionsParser.getOptionalBooleanValue(true);
@@ -182,15 +204,31 @@ public class ApkSignerTool {
                 v4SigningFlagFound = true;
             } else if ("force-stamp-overwrite".equals(optionName)) {
                 forceSourceStampOverwrite = optionsParser.getOptionalBooleanValue(true);
+            } else if ("stamp-timestamp-enabled".equals(optionName)) {
+                sourceStampTimestampEnabled = optionsParser.getOptionalBooleanValue(true);
+            } else if ("align-file-size".equals(optionName)) {
+                alignFileSize = true;
             } else if ("verity-enabled".equals(optionName)) {
                 verityEnabled = optionsParser.getOptionalBooleanValue(true);
             } else if ("debuggable-apk-permitted".equals(optionName)) {
                 debuggableApkPermitted = optionsParser.getOptionalBooleanValue(true);
+            } else if ("alignment-preserved".equals(optionName)) {
+                alignmentPreserved = optionsParser.getOptionalBooleanValue(true);
+            } else if ("lib-page-alignment".equals(optionName)) {
+                libPageAlignment = optionsParser.getRequiredIntValue(
+                        "Native library page alignment size in bytes");
             } else if ("next-signer".equals(optionName)) {
                 if (!signerParams.isEmpty()) {
                     signers.add(signerParams);
                     signerParams = new SignerParams();
                 }
+            } else if ("signer-for-min-sdk-version".equals(optionName)) {
+                if (!signerParams.isEmpty()) {
+                    signers.add(signerParams);
+                    signerParams = new SignerParams();
+                }
+                signerParams.setMinSdkVersion(optionsParser.getRequiredIntValue(
+                        "Mininimum API Level for signing config"));
             } else if ("ks".equals(optionName)) {
                 signerParams.setKeystoreFile(optionsParser.getRequiredValue("KeyStore file"));
             } else if ("ks-key-alias".equals(optionName)) {
@@ -231,8 +269,12 @@ public class ApkSignerTool {
                 signerParams.setKeyFile(optionsParser.getRequiredValue("Private key file"));
             } else if ("cert".equals(optionName)) {
                 signerParams.setCertFile(optionsParser.getRequiredValue("Certificate file"));
+            } else if ("signer-lineage".equals(optionName)) {
+                File lineageFile = new File(
+                        optionsParser.getRequiredValue("Lineage file for signing config"));
+                signerParams.setSigningCertificateLineage(getLineageFromInputFile(lineageFile));
             } else if ("lineage".equals(optionName)) {
-                File lineageFile = new File(optionsParser.getRequiredValue("Lineage File"));
+                File lineageFile = new File(optionsParser.getRequiredValue("Lineage file"));
                 lineage = getLineageFromInputFile(lineageFile);
             } else if ("v".equals(optionName) || "verbose".equals(optionName)) {
                 verbose = optionsParser.getOptionalBooleanValue(true);
@@ -257,6 +299,10 @@ public class ApkSignerTool {
                 File stampLineageFile = new File(
                         optionsParser.getRequiredValue("Stamp Lineage File"));
                 sourceStampLineage = getLineageFromInputFile(stampLineageFile);
+            } else if ("deterministic-dsa-signing".equals(optionName)) {
+                deterministicDsaSigning = optionsParser.getOptionalBooleanValue(false);
+            } else if ("append-signature".equals(optionName)) {
+                otherSignersSignaturesPreserved = optionsParser.getOptionalBooleanValue(true);
             } else {
                 throw new ParameterException(
                         "Unsupported option: " + optionOriginalForm + ". See --help for supported"
@@ -313,7 +359,8 @@ public class ApkSignerTool {
             for (SignerParams signer : signers) {
                 signerNumber++;
                 signer.setName("signer #" + signerNumber);
-                ApkSigner.SignerConfig signerConfig = getSignerConfig(signer, passwordRetriever);
+                ApkSigner.SignerConfig signerConfig = getSignerConfig(signer, passwordRetriever,
+                        deterministicDsaSigning);
                 if (signerConfig == null) {
                     return;
                 }
@@ -322,7 +369,8 @@ public class ApkSignerTool {
             if (sourceStampFlagFound) {
                 sourceStampSignerParams.setName("stamp signer");
                 sourceStampSignerConfig =
-                        getSignerConfig(sourceStampSignerParams, passwordRetriever);
+                        getSignerConfig(sourceStampSignerParams, passwordRetriever,
+                                deterministicDsaSigning);
                 if (sourceStampSignerConfig == null) {
                     return;
                 }
@@ -343,16 +391,22 @@ public class ApkSignerTool {
                 new ApkSigner.Builder(signerConfigs)
                         .setInputApk(inputApk)
                         .setOutputApk(tmpOutputApk)
-                        .setOtherSignersSignaturesPreserved(false)
+                        .setOtherSignersSignaturesPreserved(otherSignersSignaturesPreserved)
                         .setV1SigningEnabled(v1SigningEnabled)
                         .setV2SigningEnabled(v2SigningEnabled)
                         .setV3SigningEnabled(v3SigningEnabled)
                         .setV4SigningEnabled(v4SigningEnabled)
                         .setForceSourceStampOverwrite(forceSourceStampOverwrite)
+                        .setSourceStampTimestampEnabled(sourceStampTimestampEnabled)
+                        .setAlignFileSize(alignFileSize)
                         .setVerityEnabled(verityEnabled)
                         .setV4ErrorReportingEnabled(v4SigningEnabled && v4SigningFlagFound)
                         .setDebuggableApkPermitted(debuggableApkPermitted)
-                        .setSigningCertificateLineage(lineage);
+                        .setSigningCertificateLineage(lineage)
+                        .setMinSdkVersionForRotation(rotationMinSdkVersion)
+                        .setRotationTargetsDevRelease(rotationTargetsDevRelease)
+                        .setAlignmentPreserved(alignmentPreserved)
+                        .setLibraryPageAlignmentBytes(libPageAlignment);
         if (minSdkVersionSpecified) {
             apkSignerBuilder.setMinSdkVersion(minSdkVersion);
         }
@@ -389,8 +443,8 @@ public class ApkSignerTool {
         }
     }
 
-    private static ApkSigner.SignerConfig getSignerConfig(
-            SignerParams signer, PasswordRetriever passwordRetriever) {
+    private static ApkSigner.SignerConfig getSignerConfig(SignerParams signer,
+            PasswordRetriever passwordRetriever, boolean deterministicDsaSigning) {
         try {
             signer.loadPrivateKeyAndCerts(passwordRetriever);
         } catch (ParameterException e) {
@@ -420,10 +474,19 @@ public class ApkSignerTool {
         } else {
             throw new RuntimeException("Neither KeyStore key alias nor private key file available");
         }
-        ApkSigner.SignerConfig signerConfig =
+        ApkSigner.SignerConfig.Builder signerConfigBuilder =
                 new ApkSigner.SignerConfig.Builder(
-                        v1SigBasename, signer.getPrivateKey(), signer.getCerts())
-                        .build();
+                        v1SigBasename,
+                        signer.getKeyConfig(),
+                        signer.getCerts(),
+                        deterministicDsaSigning);
+        SigningCertificateLineage lineage = signer.getSigningCertificateLineage();
+        int minSdkVersion = signer.getMinSdkVersion();
+        if (minSdkVersion > 0) {
+            signerConfigBuilder.setLineageForMinSdkVersion(lineage, minSdkVersion);
+        }
+        ApkSigner.SignerConfig signerConfig = signerConfigBuilder.build();
+
         return signerConfig;
     }
 
@@ -439,6 +502,7 @@ public class ApkSignerTool {
         int maxSdkVersion = Integer.MAX_VALUE;
         boolean maxSdkVersionSpecified = false;
         boolean printCerts = false;
+        boolean printCertsPem = false;
         boolean verbose = false;
         boolean warningsTreatedAsErrors = false;
         boolean verifySourceStamp = false;
@@ -457,6 +521,13 @@ public class ApkSignerTool {
                 maxSdkVersionSpecified = true;
             } else if ("print-certs".equals(optionName)) {
                 printCerts = optionsParser.getOptionalBooleanValue(true);
+            } else if ("print-certs-pem".equals(optionName)) {
+                printCertsPem = optionsParser.getOptionalBooleanValue(true);
+                // If the PEM output of the certs is requested, this implicitly implies the
+                // cert details should be printed.
+                if (printCertsPem && !printCerts) {
+                    printCerts = true;
+                }
             } else if (("v".equals(optionName)) || ("verbose".equals(optionName))) {
                 verbose = optionsParser.getOptionalBooleanValue(true);
             } else if ("Werr".equals(optionName)) {
@@ -556,6 +627,9 @@ public class ApkSignerTool {
                 System.out.println(
                         "Verified using v3 scheme (APK Signature Scheme v3): "
                                 + result.isVerifiedUsingV3Scheme());
+                System.out.println(
+                        "Verified using v3.1 scheme (APK Signature Scheme v3.1): "
+                                + result.isVerifiedUsingV31Scheme());
                 System.out.println(
                         "Verified using v4 scheme (APK Signature Scheme v4): "
                                 + result.isVerifiedUsingV4Scheme());
@@ -565,14 +639,37 @@ public class ApkSignerTool {
                 }
             }
             if (printCerts) {
-                int signerNumber = 0;
-                for (X509Certificate signerCert : signerCerts) {
-                    signerNumber++;
-                    printCertificate(signerCert, "Signer #" + signerNumber, verbose);
+                // The v3.1 signature scheme allows key rotation to target T+ while the original
+                // signing key can still be used with v3.0; if a v3.1 block is present then also
+                // include the target SDK versions for both rotation and the original signing key.
+                if (result.isVerifiedUsingV31Scheme()) {
+                    for (ApkVerifier.Result.V3SchemeSignerInfo signer :
+                            result.getV31SchemeSigners()) {
+
+                        printCertificate(signer.getCertificate(),
+                                "Signer (minSdkVersion=" + signer.getMinSdkVersion()
+                                        + (signer.getRotationTargetsDevRelease()
+                                        ? " (dev release=true)" : "")
+                                        + ", maxSdkVersion=" + signer.getMaxSdkVersion() + ")",
+                                verbose, printCertsPem);
+                    }
+                    for (ApkVerifier.Result.V3SchemeSignerInfo signer : result.getV3SchemeSigners()) {
+                        printCertificate(signer.getCertificate(),
+                                "Signer (minSdkVersion=" + signer.getMinSdkVersion()
+                                        + ", maxSdkVersion=" + signer.getMaxSdkVersion() + ")",
+                                verbose, printCertsPem);
+                    }
+                } else {
+                    int signerNumber = 0;
+                    for (X509Certificate signerCert : signerCerts) {
+                        signerNumber++;
+                        printCertificate(signerCert, "Signer #" + signerNumber, verbose,
+                                printCertsPem);
+                    }
                 }
                 if (sourceStampInfo != null) {
                     printCertificate(sourceStampInfo.getCertificate(), "Source Stamp Signer",
-                            verbose);
+                            verbose, printCertsPem);
                 }
             }
         } else {
@@ -623,6 +720,20 @@ public class ApkSignerTool {
                         "WARNING: APK Signature Scheme v3 " + signerName + ": " + warning);
             }
         }
+        for (ApkVerifier.Result.V3SchemeSignerInfo signer : result.getV31SchemeSigners()) {
+            String signerName = "signer #" + (signer.getIndex() + 1) + "(minSdkVersion="
+                    + signer.getMinSdkVersion() + ", maxSdkVersion=" + signer.getMaxSdkVersion()
+                    + ")";
+            for (ApkVerifier.IssueWithParams error : signer.getErrors()) {
+                System.err.println(
+                        "ERROR: APK Signature Scheme v3.1 " + signerName + ": " + error);
+            }
+            for (ApkVerifier.IssueWithParams warning : signer.getWarnings()) {
+                warningsEncountered = true;
+                warningsOut.println(
+                        "WARNING: APK Signature Scheme v3.1 " + signerName + ": " + warning);
+            }
+        }
 
         if (sourceStampInfo != null) {
             for (ApkVerifier.IssueWithParams error : sourceStampInfo.getErrors()) {
@@ -631,6 +742,9 @@ public class ApkSignerTool {
             for (ApkVerifier.IssueWithParams warning : sourceStampInfo.getWarnings()) {
                 warningsOut.println("WARNING: SourceStamp: " + warning);
             }
+            for (ApkVerifier.IssueWithParams infoMessage : sourceStampInfo.getInfoMessages()) {
+                System.out.println("INFO: SourceStamp: " + infoMessage);
+            }
         }
 
         if (!verified) {
@@ -732,7 +846,8 @@ public class ApkSignerTool {
             loadPrivateKeyAndCerts(oldSignerParams, passwordRetriever);
             SigningCertificateLineage.SignerConfig oldSignerConfig =
                     new SigningCertificateLineage.SignerConfig.Builder(
-                            oldSignerParams.getPrivateKey(), oldSignerParams.getCerts().get(0))
+                                    oldSignerParams.getKeyConfig(),
+                                    oldSignerParams.getCerts().get(0))
                             .build();
 
             // TOOD: don't require private key
@@ -740,7 +855,8 @@ public class ApkSignerTool {
             loadPrivateKeyAndCerts(newSignerParams, passwordRetriever);
             SigningCertificateLineage.SignerConfig newSignerConfig =
                     new SigningCertificateLineage.SignerConfig.Builder(
-                            newSignerParams.getPrivateKey(), newSignerParams.getCerts().get(0))
+                                    newSignerParams.getKeyConfig(),
+                                    newSignerParams.getCerts().get(0))
                             .build();
 
             // ok we're all set up, let's rotate!
@@ -783,6 +899,7 @@ public class ApkSignerTool {
 
         boolean verbose = false;
         boolean printCerts = false;
+        boolean printCertsPem = false;
         boolean lineageUpdated = false;
         File inputKeyLineage = null;
         File outputKeyLineage = null;
@@ -804,6 +921,13 @@ public class ApkSignerTool {
                 verbose = optionsParser.getOptionalBooleanValue(true);
             } else if ("print-certs".equals(optionName)) {
                 printCerts = optionsParser.getOptionalBooleanValue(true);
+            } else if ("print-certs-pem".equals(optionName)) {
+                printCertsPem = optionsParser.getOptionalBooleanValue(true);
+                // If the PEM output of the certs is requested, this implicitly implies the
+                // cert details should be printed.
+                if (printCertsPem && !printCerts) {
+                    printCerts = true;
+                }
             } else {
                 throw new ParameterException(
                         "Unsupported option: " + optionsParser.getOptionOriginalForm()
@@ -822,7 +946,7 @@ public class ApkSignerTool {
                 loadPrivateKeyAndCerts(signerParams, passwordRetriever);
                 SigningCertificateLineage.SignerConfig signerConfig =
                         new SigningCertificateLineage.SignerConfig.Builder(
-                                signerParams.getPrivateKey(), signerParams.getCerts().get(0))
+                                        signerParams.getKeyConfig(), signerParams.getCerts().get(0))
                                 .build();
                 try {
                     // since only the caller specified capabilities will be updated a direct
@@ -864,7 +988,8 @@ public class ApkSignerTool {
             for (int i = 0; i < signingCerts.size(); i++) {
                 X509Certificate signerCert = signingCerts.get(i);
                 SignerCapabilities signerCapabilities = lineage.getSignerCapabilities(signerCert);
-                printCertificate(signerCert, "Signer #" + (i + 1) + " in lineage", verbose);
+                printCertificate(signerCert, "Signer #" + (i + 1) + " in lineage", verbose,
+                        printCertsPem);
                 printCapabilities(signerCapabilities);
             }
         }
@@ -995,6 +1120,14 @@ public class ApkSignerTool {
         }
     }
 
+    /**
+     * @see #printCertificate(X509Certificate, String, boolean, boolean)
+     */
+    public static void printCertificate(X509Certificate cert, String name, boolean verbose)
+            throws NoSuchAlgorithmException, CertificateEncodingException {
+        printCertificate(cert, name, verbose, false);
+    }
+
     /**
      * Prints details from the provided certificate to stdout.
      *
@@ -1002,13 +1135,15 @@ public class ApkSignerTool {
      * @param name    the name to be used to identify the certificate.
      * @param verbose boolean indicating whether public key details from the certificate should be
      *                displayed.
+     * @param pemOutput boolean indicating whether the PEM encoding of the certificate should be
+     *                  displayed.
      * @throws NoSuchAlgorithmException     if an instance of MD5, SHA-1, or SHA-256 cannot be
      *                                      obtained.
      * @throws CertificateEncodingException if an error is encountered when encoding the
      *                                      certificate.
      */
-    public static void printCertificate(X509Certificate cert, String name, boolean verbose)
-            throws NoSuchAlgorithmException, CertificateEncodingException {
+    public static void printCertificate(X509Certificate cert, String name, boolean verbose,
+            boolean pemOutput) throws NoSuchAlgorithmException, CertificateEncodingException {
         if (cert == null) {
             throw new NullPointerException("cert == null");
         }
@@ -1053,6 +1188,18 @@ public class ApkSignerTool {
             System.out.println(
                     name + " public key MD5 digest: " + HexEncoding.encode(md5.digest(encodedKey)));
         }
+
+        if (pemOutput) {
+            System.out.println(BEGIN_CERTIFICATE);
+            final int lineWidth = 64;
+            String pemEncodedCert = Base64.getEncoder().encodeToString(cert.getEncoded());
+            for (int i = 0; i < pemEncodedCert.length(); i += lineWidth) {
+                System.out.println(pemEncodedCert.substring(i, i + lineWidth > pemEncodedCert.length()
+                        ? pemEncodedCert.length()
+                        : i + lineWidth));
+            }
+            System.out.println(END_CERTIFICATE);
+        }
     }
 
     /**
