@@ -16,9 +16,13 @@
 
 package com.android.apksig.internal.apk.v1;
 
+import static com.android.apksig.Constants.MAX_APK_SIGNERS;
+import static com.android.apksig.Constants.OID_RSA_ENCRYPTION;
 import static com.android.apksig.internal.pkcs7.AlgorithmIdentifier.getSignerInfoDigestAlgorithmOid;
 import static com.android.apksig.internal.pkcs7.AlgorithmIdentifier.getSignerInfoSignatureAlgorithm;
 
+import com.android.apksig.KeyConfig;
+import com.android.apksig.SignerEngineFactory;
 import com.android.apksig.apk.ApkFormatException;
 import com.android.apksig.internal.apk.ApkSigningBlockUtils;
 import com.android.apksig.internal.asn1.Asn1EncodingException;
@@ -30,6 +34,7 @@ import com.android.apksig.internal.util.Pair;
 import java.io.ByteArrayInputStream;
 import java.io.ByteArrayOutputStream;
 import java.io.IOException;
+import java.security.InvalidAlgorithmParameterException;
 import java.security.InvalidKeyException;
 import java.security.MessageDigest;
 import java.security.NoSuchAlgorithmException;
@@ -76,12 +81,21 @@ public abstract class V1SchemeSigner {
         /** Name. */
         public String name;
 
-        /** Private key. */
+        /**
+         * Private key.
+         *
+         * @deprecated all internal usage has migrated to use {@link #keyConfig}. This field is not
+         *     removed so that compilation is not broken for clients referencing it, but using this
+         *     field may lead to unexpected errors.
+         */
         public PrivateKey privateKey;
 
+        /** Signing key configuration */
+        public KeyConfig keyConfig;
+
         /**
          * Certificates, with the first certificate containing the public key corresponding to
-         * {@link #privateKey}.
+         * {@link #keyConfig}.
          */
         public List<X509Certificate> certificates;
 
@@ -89,6 +103,11 @@ public abstract class V1SchemeSigner {
          * Digest algorithm used for the signature.
          */
         public DigestAlgorithm signatureDigestAlgorithm;
+
+        /**
+         * If DSA is the signing algorithm, whether or not deterministic DSA signing should be used.
+         */
+        public boolean deterministicDsaSigning;
     }
 
     /** Hidden constructor to prevent instantiation. */
@@ -106,7 +125,7 @@ public abstract class V1SchemeSigner {
     public static DigestAlgorithm getSuggestedSignatureDigestAlgorithm(
             PublicKey signingKey, int minSdkVersion) throws InvalidKeyException {
         String keyAlgorithm = signingKey.getAlgorithm();
-        if ("RSA".equalsIgnoreCase(keyAlgorithm)) {
+        if ("RSA".equalsIgnoreCase(keyAlgorithm) || OID_RSA_ENCRYPTION.equals((keyAlgorithm))) {
             // Prior to API Level 18, only SHA-1 can be used with RSA.
             if (minSdkVersion < 18) {
                 return DigestAlgorithm.SHA1;
@@ -241,6 +260,11 @@ public abstract class V1SchemeSigner {
         if (signerConfigs.isEmpty()) {
             throw new IllegalArgumentException("At least one signer config must be provided");
         }
+        if (signerConfigs.size() > MAX_APK_SIGNERS) {
+            throw new IllegalArgumentException(
+                    "APK Signature Scheme v1 only supports a maximum of " + MAX_APK_SIGNERS + ", "
+                            + signerConfigs.size() + " provided");
+        }
         OutputManifestFile manifest =
                 generateManifestFile(
                         jarEntryDigestAlgorithm, jarEntryDigests, sourceManifestBytes);
@@ -495,19 +519,20 @@ public abstract class V1SchemeSigner {
         PublicKey publicKey = signingCert.getPublicKey();
         DigestAlgorithm digestAlgorithm = signerConfig.signatureDigestAlgorithm;
         Pair<String, AlgorithmIdentifier> signatureAlgs =
-                getSignerInfoSignatureAlgorithm(publicKey, digestAlgorithm);
+                getSignerInfoSignatureAlgorithm(publicKey, digestAlgorithm,
+                        signerConfig.deterministicDsaSigning);
         String jcaSignatureAlgorithm = signatureAlgs.getFirst();
 
         // Generate the cryptographic signature of the signature file
         byte[] signatureBytes;
         try {
-            Signature signature = Signature.getInstance(jcaSignatureAlgorithm);
-            signature.initSign(signerConfig.privateKey);
-            signature.update(signatureFileBytes);
-            signatureBytes = signature.sign();
+            signatureBytes =
+                    SignerEngineFactory.getImplementation(
+                                    signerConfig.keyConfig, jcaSignatureAlgorithm, null)
+                            .sign(signatureFileBytes);
         } catch (InvalidKeyException e) {
             throw new InvalidKeyException("Failed to sign using " + jcaSignatureAlgorithm, e);
-        } catch (SignatureException e) {
+        } catch (InvalidAlgorithmParameterException | SignatureException e) {
             throw new SignatureException("Failed to sign using " + jcaSignatureAlgorithm, e);
         }
 
